/*
  Copyright (C) Information Equipment co.,LTD
  All rights reserved.
  Code by JaeHyuk Cho <mailto:minzkn@minzkn.com>
  CVSTAG="$Id: cpuusage.c,v 1.3 2008/06/11 02:02:50 minzkn Exp $"
*/

#if !defined(DEF_SOURCE_cpuusage_c)
# define DEF_SOURCE_cpuusage_c "cpuusage.c"

#include <sys/time.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include "cpuusage.h"

#define DEF_MAX_CPU (16)

int MZ_GetCPU_StatBuffer(void *s_Buffer, int s_Length);
int MZ_GetCPU_Parse(char s_How, void *s_Buffer, t_CPU_Tick *s_CPU_Tick);
void MZ_GetCPU_MakeSubInfo(t_CPU_Tick *s_SubInfo, const t_CPU_Tick *s_OldInfo, const t_CPU_Tick *s_NewInfo);

int MZ_GetCPU_StatBuffer(void *s_Buffer, int s_Length)
{
 const char c_Proc_CPU_Stat[] = "/proc/stat";	
 int s_Handle, s_ReadBytes;
 s_Handle = open(c_Proc_CPU_Stat, O_RDONLY); 
 if(s_Handle >= 0)
 {
  s_ReadBytes = read(s_Handle, s_Buffer, s_Length - 1);
  close(s_Handle);
  if(s_ReadBytes > 0)*(((char *)s_Buffer) + s_ReadBytes) = '\0';	 
 }
 else s_ReadBytes = (-1);
 return(s_ReadBytes); 
}

int MZ_GetCPU_Parse(char s_How, void *s_Buffer, t_CPU_Tick *s_CPU_Tick)
{
 char s_HowCPU[ 3 + 3 + 1 ], *s_CPU_Entry;
 int s_Return = 0, s_Check;
 sprintf((char *)(&s_HowCPU[0]), "cpu%c", s_How);
 s_CPU_Entry = strstr((char *)s_Buffer, (char *)(&s_HowCPU[0]));
 if(s_CPU_Entry)
 {
  s_CPU_Tick->Ex[0] = 0ul; /* iowait */
  s_CPU_Tick->Ex[1] = 0ul; /* irq */
  s_CPU_Tick->Ex[2] = 0ul; /* softirq */
  s_CPU_Tick->Ex[3] = 0ul; /* stolen */
  s_Check = sscanf(s_CPU_Entry, "%s %lu %lu %lu %lu %lu %lu %lu %lu", (char *)(&s_HowCPU[0]),
                   (unsigned long *)(&s_CPU_Tick->User), (unsigned long *)(&s_CPU_Tick->Nice),
                   (unsigned long *)(&s_CPU_Tick->System), (unsigned long *)(&s_CPU_Tick->Idle),
		   (unsigned long *)(&s_CPU_Tick->Ex[0]),
		   (unsigned long *)(&s_CPU_Tick->Ex[1]),
		   (unsigned long *)(&s_CPU_Tick->Ex[2]),
		   (unsigned long *)(&s_CPU_Tick->Ex[3]));
  if(s_Check >= (5 /* + 4 */))
  {
   s_CPU_Tick->How = s_How;	  
   s_CPU_Tick->Total = s_CPU_Tick->User + s_CPU_Tick->Nice + s_CPU_Tick->System + s_CPU_Tick->Idle + s_CPU_Tick->Ex[0] + s_CPU_Tick->Ex[1] + s_CPU_Tick->Ex[2] + s_CPU_Tick->Ex[3];
   s_Return = 1;
  }
 }
 return(s_Return); 
}

void MZ_GetCPU_MakeSubInfo(t_CPU_Tick *s_SubInfo, const t_CPU_Tick *s_OldInfo, const t_CPU_Tick *s_NewInfo)
{
 s_SubInfo->How = s_NewInfo->How;
 #define __DEF_CPU_SUBINFO__(this) do{s_SubInfo->this = s_NewInfo->this - s_OldInfo->this;}while(0) 
 __DEF_CPU_SUBINFO__(User);   __DEF_CPU_SUBINFO__(Nice);
 __DEF_CPU_SUBINFO__(System); __DEF_CPU_SUBINFO__(Idle);
 __DEF_CPU_SUBINFO__(Ex[0]); __DEF_CPU_SUBINFO__(Ex[1]);
 __DEF_CPU_SUBINFO__(Ex[2]); __DEF_CPU_SUBINFO__(Ex[3]);
 __DEF_CPU_SUBINFO__(Total);
 #undef __DEF_CPU_SUBINFO__
 memcpy((void *)s_OldInfo, (void *)s_NewInfo, sizeof(t_CPU_Tick));
}

static void MZ_GetCPU_Print(t_CPU_Tick *s_SubInfo, int s_Mode, unsigned long long s_CPU_Clock)
{
 char s_CPU_Name[ 1 << 4 ];
 int s_MaxWidth = 40, s_LeftWidth = 0, s_Count;
 if(s_SubInfo->Total <= 0)s_SubInfo->Total = 1lu;
 if(s_Mode == 0)
 { 
  s_LeftWidth = ((s_SubInfo->Total - s_SubInfo->Idle) * s_MaxWidth) / s_SubInfo->Total;
  s_SubInfo->Idle = ((s_SubInfo->Total - s_SubInfo->Idle)   * 10000lu) / s_SubInfo->Total;
 }
 else if(s_Mode == 1)
 {
  s_SubInfo->User   = (s_SubInfo->User   * 10000lu) / s_SubInfo->Total;
  s_SubInfo->Nice   = (s_SubInfo->Nice   * 10000lu) / s_SubInfo->Total;
  s_SubInfo->System = (s_SubInfo->System * 10000lu) / s_SubInfo->Total;
  s_SubInfo->Idle   = (s_SubInfo->Idle   * 10000lu) / s_SubInfo->Total;
  s_SubInfo->Ex[0]  = (s_SubInfo->Ex[0]  * 10000lu) / s_SubInfo->Total;
  s_SubInfo->Ex[1]  = (s_SubInfo->Ex[1]  * 10000lu) / s_SubInfo->Total;
  s_SubInfo->Ex[2]  = (s_SubInfo->Ex[2]  * 10000lu) / s_SubInfo->Total;
  s_SubInfo->Ex[3]  = (s_SubInfo->Ex[3]  * 10000lu) / s_SubInfo->Total;
 }
 if(s_SubInfo->How == ' ')sprintf((char *)(&s_CPU_Name[0]), "CPU");
 else sprintf((char *)(&s_CPU_Name[0]), "CPU #%c", s_SubInfo->How);
 if(s_Mode == 0)
 {
  fprintf(stdout, "[%-6s] [", (char *)(&s_CPU_Name[0]));
  for(s_Count = 0;s_Count < s_LeftWidth;s_Count++)fprintf(stdout, "=");
  while((s_Count++) < s_MaxWidth)fprintf(stdout, " ");
  if(s_SubInfo->How == ' ')
  {
   fprintf(stdout, "] %3lu.%02lu%% - %llu.%03lluGHz\n", 
		   s_SubInfo->Idle / 100lu, s_SubInfo->Idle % 100lu, 
		   s_CPU_Clock / 1000000000llu, (s_CPU_Clock % 1000000000llu) / 1000000llu); 
  }
  else
  {
   fprintf(stdout, "] %3lu.%02lu%%\n", 
		   s_SubInfo->Idle / 100lu, s_SubInfo->Idle % 100lu);

  }
 }
 else if(s_Mode == 1)
 {
  if(s_SubInfo->How == ' ')
  {
   fprintf(stdout, "[%-6s] %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% - %llu.%03lluGHz\n",
           (char *)(&s_CPU_Name[0]), 
           s_SubInfo->User   / 100lu, s_SubInfo->User   % 100lu,
           s_SubInfo->Nice   / 100lu, s_SubInfo->Nice   % 100lu,
           s_SubInfo->System / 100lu, s_SubInfo->System % 100lu,
           s_SubInfo->Idle   / 100lu, s_SubInfo->Idle   % 100lu,
           s_SubInfo->Ex[0]  / 100lu, s_SubInfo->Ex[0]  % 100lu,
           s_SubInfo->Ex[1]  / 100lu, s_SubInfo->Ex[1]  % 100lu,
           s_SubInfo->Ex[2]  / 100lu, s_SubInfo->Ex[2]  % 100lu,
           s_SubInfo->Ex[3]  / 100lu, s_SubInfo->Ex[3]  % 100lu,
           s_CPU_Clock / 1000000000llu, (s_CPU_Clock % 1000000000llu) / 1000000llu); 
  }
  else
  {
   fprintf(stdout, "[%-6s] %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%% %3lu.%02lu%%\n",
           (char *)(&s_CPU_Name[0]), 
           s_SubInfo->User   / 100lu, s_SubInfo->User   % 100lu,
           s_SubInfo->Nice   / 100lu, s_SubInfo->Nice   % 100lu,
           s_SubInfo->System / 100lu, s_SubInfo->System % 100lu,
           s_SubInfo->Idle   / 100lu, s_SubInfo->Idle   % 100lu,
           s_SubInfo->Ex[0]  / 100lu, s_SubInfo->Ex[0]  % 100lu,
           s_SubInfo->Ex[1]  / 100lu, s_SubInfo->Ex[1]  % 100lu,
           s_SubInfo->Ex[2]  / 100lu, s_SubInfo->Ex[2]  % 100lu,
           s_SubInfo->Ex[3]  / 100lu, s_SubInfo->Ex[3]  % 100lu);
  }
 }
}

static unsigned long long MZ_GetCPU_Clock(void)
{
 register unsigned long long s_Return;	
#if defined(__i386__) /* i386 only */
 __asm__ volatile("rdtsc\n\t":"=A"(s_Return):);
#else
 s_Return = (unsigned long long)0;
#endif
 return(s_Return);
}

static void MZ_Sleep(int s_Sec, int s_uSec)
{
#if 0
 if(s_Sec > 0)sleep(s_Sec);
 if(s_uSec > 0)usleep(s_uSec);
#else
 struct timeval s_TimeVal;
 s_TimeVal.tv_sec = s_Sec;
 s_TimeVal.tv_usec = s_uSec;
 (void)select(0, (fd_set *)0, (fd_set *)0, (fd_set *)0, &s_TimeVal);
#endif
}

static int __g_Ctrl_C__ = 0;

static void __MZ_SIGINT_SIGTERM__(int s_Signal)
{
 __g_Ctrl_C__ = 1;
}

int main(int s_Argc, char *s_Argv[])
{
 t_CPU_Tick s_OldInfo[DEF_MAX_CPU], s_NewInfo[DEF_MAX_CPU], s_SubInfo[DEF_MAX_CPU];
 char s_Buffer[ 4 << 10 ];
 int s_CPU_Index, s_MaxCPU = 1;
 int s_PrintMode = 0, s_ArgCount;
 unsigned long long s_CPU_Clock = (unsigned long long)0;
 
 signal(SIGINT, __MZ_SIGINT_SIGTERM__);
 signal(SIGTERM, __MZ_SIGINT_SIGTERM__);

 fprintf(stdout, "cpuusage v1.0.1 - %s %s\n"
		 "Code by JaeHyuk Cho <mailto:minzkn@minzkn.com> , Application name : cpuusage , Made in KOREA.\n\n"
		 "Usage: %s [detail | all]\n\n", __DATE__, __TIME__, s_Argv[0]);
 if(s_Argc > 1)
 {
  for(s_ArgCount = 1;s_ArgCount < s_Argc;s_ArgCount++)
  {	  
   if(strcmp(s_Argv[s_ArgCount], "detail") == 0)s_PrintMode = 1;
   else if(strcmp(s_Argv[s_ArgCount], "all") == 0)s_MaxCPU = DEF_MAX_CPU; 
  }
 } 
 memset((void *)(&s_OldInfo[0]), 0, sizeof(s_OldInfo));
 do
 {
  if(s_MaxCPU > 1)fprintf(stdout, "----------------------------------------\n");
  if(MZ_GetCPU_StatBuffer((void *)(&s_Buffer[0]), sizeof(s_Buffer)) > 0)
  {
   s_CPU_Index = 0;
   while(MZ_GetCPU_Parse(s_CPU_Index == 0 ? ' ' : ((s_CPU_Index - 1) + '0'), (void *)(&s_Buffer[0]), &s_NewInfo[s_CPU_Index]) > 0)
   {
    MZ_GetCPU_MakeSubInfo(&s_SubInfo[s_CPU_Index], &s_OldInfo[s_CPU_Index], &s_NewInfo[s_CPU_Index]);    
    MZ_GetCPU_Print(&s_SubInfo[s_CPU_Index], s_PrintMode, s_CPU_Clock);
    if((++s_CPU_Index) >= s_MaxCPU)break;
   }
  }
  s_CPU_Clock = MZ_GetCPU_Clock();
  MZ_Sleep(1, 0);
  s_CPU_Clock = (MZ_GetCPU_Clock() - s_CPU_Clock);
 }while(__g_Ctrl_C__ == 0);
 fprintf(stdout, "\nEnd of cpuusage.\n");
 return(0);
}

#endif

/* End of source */
